import base
import cache
import config
import fileinfo
import time
import utils

def get_instance():
	return Autocommit.get_instance()

class Autocommit(base.Singleton):
	def __init__(self):
		self._config = config.get_instance()
		if not self.is_enabled():
			return

		self._auto_soft = 60 * self._config.getint("autocommit.soft", 20)
		self._auto_hard = 60 * self._config.getint("autocommit.hard", 1440)

		self._cache = cache.Cache("autocommit")
		if self._cache.data is None:
			self._cache.data = { "stall" : set(), "stall_all" : False }

	def is_enabled(self):
		return self._config.getboolean("autocommit.enabled", True)

	def time(self, changeset):
		"""Get time of an autocommit for a given ChangeSet."""
		assert type(changeset) is fileinfo.changes.ChangeSet

		time_first = changeset.time_first_change()
		time_last = changeset.time_last_change()

		time_soft = time_last + self._auto_soft
		time_hard = time_first + self._auto_hard

		if time_hard < time_soft:
			return time_hard
		else:
			return time_soft

	def message(self, changeset, package):
		"""Get a message describing time of next autocommit for a given changeset."""
		assert type(changeset) is fileinfo.changes.ChangeSet

		if not self.is_enabled():
			return "disabled"

		if self.is_stalled(package):
			return "stalled"

		time_auto = self.time(changeset)
		remains = utils.time_remains(time_auto)

		if remains <= 0:
			return "very soon"
		else:
			return utils.time_relative(time_auto)

	def is_stalled(self, package):
		return self.is_stalled_all() or package in self._cache.data["stall"]

	def is_stalled_all(self):
		return self._cache.data["stall_all"]

	def set_stall(self, package):
		self._cache.data["stall"].add(package)
		self._cache.save()

	def set_auto(self, package):
		self._cache.data["stall"].remove(package)
		self._cache.save()

	def set_stall_all(self):
		self._cache.data["stall_all"] = True
		self._cache.save()

	def set_auto_all(self):
		self._cache.data["stall_all"] = False
		self._cache.data["stall"] = set()
		self._cache.save()
