import os
import pickle
import utils

__all__ = [ "Cache", "set_location" ]

DEFAULT_CACHE_LOCATION = "/var/lib/sandglass/cache"
PICKLE_PROTOCOL = pickle.HIGHEST_PROTOCOL

def set_location(path):
	Cache.set_location(path)

class Cache(object):
	"""
	Universal class for simple object storing on local filesystem.
	Can be used as database or cache. Uses Python pickle as internal storage.
	"""

	_path = DEFAULT_CACHE_LOCATION
	@classmethod
	def set_location(cls, path):
		"""
		Set location for cache files.
		"""
		cls._path = path

	def __init__(self, name, public = True):
		"""
		Create new cache object.

		Argument 'name' specifies filename (relative to cache localtion, see set_location) to use.

		If 'public' is set to True, cache file will be readable by everyone (0644). Otherwise only owner
		will be able to use the cache file (0600).
		"""
		assert Cache._path
		self._filename = os.path.join(Cache._path, name)
		self._public = public
		self._initialized = False
		self._data = None

		parent = os.path.dirname(self._filename)
		if not os.path.isdir(parent):
			os.makedirs(parent, 0755)

	@property
	def data(self):
		"""
		Get object stored in cache (shallow copy).

		Note that data structures obtained using this method are not saved to cache after any
		modifications. To save the state of the object to cache, use cache.data = data
		or cache.save().
		"""
		if not self._initialized:
			self._load()
			self._initialized = True

		return self._data

	@data.setter
	def data(self, value):
		"""
		Save new data to cache.
		"""
		self._data = value
		self._save()
		self._initialized = True

	def save(self):
		"""
		Save the object to the cache.
		"""
		self._save()

	def _load(self):
		"""
		Load the object from the filesystem.
		"""
		try:
			with open(self._filename, "rb") as f:
				self._data = pickle.load(f)
		except IOError:
			pass

	def _save(self):
		"""
		Save the object to the filesystem.
		"""
		if not os.path.exists(self._filename):
			open(self._filename, "w").close()
			os.chmod(self._filename, 0644 if self._public else 0600)
		with open(self._filename, "wb") as f:
			pickle.dump(self._data, f, PICKLE_PROTOCOL)

	def saved_at(self):
		"""
		Return timestamp of the last save operation.
		"""
		return utils.file_mtime(self._filename, must_exist=False)
