import repository
import override
import exceptions
import utils
import fileinfo
import cache
import diff
import config
import os
import shutil
import re
import git

need_initialized = repository.need_initialized
need_not_initialized = repository.need_not_initialized
SANDGLASS_SSH = "/usr/libexec/sandglass/ssh.sh"

class ClientRepository(repository.Repository):
	def __init__(self, data_dir = None, root_dir = None):
		super(ClientRepository, self).__init__()
		assert data_dir is not None
		assert root_dir is not None

		self._repo_dir = os.path.join(data_dir, "repository")
		self._root_dir = root_dir

		# some git commands behave strange (e.g. git ls-files) when CWD is not root directory
		os.chdir("/")
		self._setup()

	def _setup(self):
		if os.path.isdir(self._repo_dir):
			self._repo = override.Repo(self._repo_dir, self._root_dir)

		# set SSH client proxy script (used for authentication)
		os.environ["GIT_SSH"] = SANDGLASS_SSH

	def _get_commit(self, commit=None, implicit_head=True):
		"""
		Get git.object.Commit object from other type of commit representation (string).
		"""
		if commit is None and implicit_head:
			commit = self._repo.active_branch.commit
		return super(ClientRepository, self)._get_commit(commit)

	def get_attributes(self, commit=None):
		"""
		Get AttributeSet with information taken from a given commit.
		"""
		if commit is None:
			commit = self._repo.active_branch.commit
		return super(ClientRepository, self).get_attributes(commit)

	def _working_tree_attributes(self):
		"""
		Get AttributeSet with fresh information about all tracked files in working directory.
		"""
		attrs = fileinfo.attributes.AttributeSet(self.tracked())
		attrs.refresh()
		return attrs

	def real_filename(self, internal_path):
		"""
		Convert name of file stored in repository to real name on filesystem.
		"""
		return os.path.join(self._root_dir, internal_path)

	def repo_filename(self, real_path):
		"""
		Opposite of real_filename(). Returns virtual repository path.
		"""
		# workaround for bug #704793 in Fedora 14
		if self._root_dir == "/" and real_path.startswith("/"):
			return real_path[1:]
		return os.path.relpath(real_path, self._root_dir)

	@need_not_initialized
	def _init_repo(self):
		"""
		Initialize repository and set all necessary configuraion.
		"""
		utils.secure_makedirs(self._repo_dir)
		new_repo = override.Repo.init(self._repo_dir, bare = True)

		hostname = utils.hostname()
		config = new_repo.config_writer()

		config.set("core", "bare", False)
		config.set("core", "worktree", self._root_dir)
		config.add_section("user")
		config.set("user", "name", "root (%s)" % hostname)
		config.set("user", "email", "root@%s" % hostname)

		# needed for remote synchronization
		if not config.has_section("push"):
			config.add_section("push")
		config.set("push", "default", "upstream")

	def init(self, force = False):
		"""
		Initialize new repository. If the repository is already initialized, 'force' has to be set to True.
		"""
		if force:
			shutil.rmtree(self._repo_dir, True)
			self._repo = None
		self._init_repo()
		self._setup()

	@need_initialized
	def diff(self, old_commit=None, new_commit=None, files=None):
		"""
		Show diff between two revisions. The output can be limited by passing a list of files.
		If 'old_commit' is not specified, use the branch head commit.
		If 'new_commit' is not specified, use working tree state (not really a commit).
		"""
		diff_helper = diff.DiffHelper(self)

		old_attrs = self.get_attributes(old_commit)
		if new_commit is not None:
			new_attrs = self.get_attributes(new_commit)
		else:
			new_attrs = self._working_tree_attributes()

		git_output = self._repo.git.diff(old_commit, new_commit, "--", files)
		sgl_output = diff_helper.git2sgl(git_output, old_attrs, new_attrs, limit_files=files)
		return sgl_output

	@need_initialized
	def dump(self, filename, commit=None):
		"""
		Get file content at a given commit. If the commit is not specified, use the last one.
		"""
		commit = self._get_commit(commit)
		filename = self.repo_filename(filename)
		return commit.tree[filename].data

	def _recover_tmpfile(self, filename):
		"""
		Get temporary file name for a recovered file.
		"""
		dirname = os.path.dirname(filename)
		basename = ".sglrecover_%s" % os.path.basename(filename)
		tmp_filename = os.path.join(dirname, basename)
		return self.real_filename(tmp_filename)

	def _recover_file(self, commit, filename, attributes):
		"""
		Safely recover one file including all attribtues.
		"""
		blob = self._get_commit(commit).tree[filename]
		filename = self.real_filename(filename)
		tmp_filename = self._recover_tmpfile(filename)

		with utils.secure_open(tmp_filename, "wb") as new_fd:
			blob.stream_data(new_fd)
		attributes.save(tmp_filename)
		os.rename(tmp_filename, filename)

	@need_initialized
	def recover(self, commit=None, files=None, all=False):
		"""
		Recover file(s) including attributes from repository. If the commit is not given,
		use the last one.
		"""
		if all:
			files = self.changed(commit).files()

		commit = self._get_commit(commit)
		attrs = self.get_attributes(commit)
		files = map(self.repo_filename, files)

		for file in files:
			commit.tree[file]
		for file in files:
			self._recover_file(commit, file, attrs[self.real_filename(file)])

	@need_initialized
	def add(self, files):
		"""
		Add files to be tracked.
		"""
		files = filter(self._allowed_file, map(self.repo_filename, files))
		if len(files) == 0:
			raise exceptions.InvalidOperation("No files found to be added.")
		self._repo.git.add(files)

	@need_initialized
	def tracked(self):
		"""
		Return list of tracked files.
		"""
		files = self._repo.git.ls_files(cached = True).splitlines()
		return map(self.real_filename, filter(self._allowed_file, files))

	def _git_changed(self, changes, commit=None):
		"""
		Update ChangeSet object by adding information about changes, which can be extracted from Git.
		This covers information about file modification and file removal.
		"""
		assert type(changes) is fileinfo.changes.ChangeSet

		# Git detects executable bits changes, --raw output doesn't give us enough information
		# to recognize whether the file was really modified or whether only the mode changed.
		# We utilize --numstat option to get this information.

		lines = self._repo.git.diff(commit, raw=True, numstat=True, diff_filter="DM").splitlines()
		# (see 'man git-diff')
		# :100644 000000 f2327cb... 0000000... D  .sglattributes
		# :100644 100644 54666cc... 0000000... M  etc/adjtime
		# :100644 100755 b2c0cfd... 0000000... M  etc/passwd
		# 0       295     .sglattributes
		# 2       2       etc/adjtime
		# 0       0       etc/passwd

		for line in lines:
			if line.startswith(":"):
				(nop, nop, nop, nop, change, filename) = line.split()
				filename = self.real_filename(filename)
				if not self._allowed_file(filename):
					continue
				if change == "D":
					operation = fileinfo.changes.Delete()
					changes.add_or_update(filename, operation)
				elif change == "M":
					# wait for the second part of the output
					pass
				else:
					raise exceptions.UnexpectedResult("Unexpected change type - '%s'." % change)
			else:
				(added, removed, filename) = line.split()
				filename = self.real_filename(filename)
				if not self._allowed_file(filename):
					continue
				added = int(added)
				removed = int(removed)

				if not (added == removed == 0):
					mtime = utils.file_mtime(filename, must_exist=False)
					operation = fileinfo.changes.Modify(added, removed, mtime)
					changes.add_or_update(filename, operation)
				else:
					# file mode changed, but we have no details
					pass

		return changes

	def _sgl_changed(self, changes, commit=None):
		"""
		Update ChangeSet object by adding additional information about file changes, which cannot
		be extracted from Git and Sandglass need to track. This means file attributes changes
		(ownership, permission, SELinux context).
		"""
		assert type(changes) is fileinfo.changes.ChangeSet

		new_attr = self._working_tree_attributes()
		old_attr = self.get_attributes(commit)
		for file in new_attr.files():
			diff = old_attr[file].diff(new_attr[file])
			if diff is not None:
				changes.add_or_update(file, diff)

	@need_initialized
	def changed(self, commit=None, drop_cache=False):
		"""
		Return ChangeSet object with information about changes performed on tracked files.
		"""
		changes = fileinfo.changes.ChangeSet()
		self._git_changed(changes, commit)
		self._sgl_changed(changes, commit)

		# use cache only when we are working with HEAD
		if commit is None:
			cached_changes = cache.Cache("changes_head")
			if not drop_cache and cached_changes.data is not None:
				changes.merge_old(cached_changes.data)
			cached_changes.data = changes

		return changes

	@need_initialized
	def untrack(self, files):
		"""
		Remove files from the list of tracked files.
		"""
		files_set = set(filter(self._allowed_file, files))
		if not files_set <= set(self.tracked()):
			raise exceptions.InvalidOperation("Supplied files are not tracked.")
		if len(files) == 0:
			raise exceptions.InvalidOperation("No files to be removed.")

		# running 'git rm --cached' always return exit-code 1, we have to verify the result of the operation
		raised = Exception()
		try:
			self._repo.git.rm(list(files), cached = True)
		except git.errors.GitCommandError as e:
			raised = e
		if len(files_set & set(self.tracked())) != 0:
			raise raised

	def _add_attributes(self):
		"""
		Add fresh .sglattributes file to index.
		"""
		file = utils.temp_file()
		attrs = fileinfo.attributes.AttributeSet(self.tracked())
		attrs.save(file)

		hash = self._repo.git.hash_object("-w", file).strip()
		self._repo.git.update_index("--add", "--cacheinfo", "0", hash, ".sglattributes")

	@need_initialized
	def commit(self, message):
		"""
		Performs a commit.
		"""
		self._add_attributes()
		self._repo.index.commit(message)

	def _remote_set_server(self, remote_name, url, branch):
		"""
		Set remote server.
		"""
		# delete old if it already exists
		for remote in self._repo.remotes:
			if remote.name == remote_name:
				self._repo.delete_remote(remote)
				break

		self._repo.create_remote(remote_name, url)

		# configuration includes:
		# - fetching only client references
		# - default references for merging

		config = self._repo.config_writer()
		section_remote = 'remote "%s"' % remote_name

		fetch = "+refs/heads/%(branch)s:refs/remotes/%(remote)s/%(branch)s" % \
			{ "branch" : branch, "remote" : remote_name }
		config.set(section_remote, "fetch", fetch)

		merge = "refs/heads/%s" % branch
		section_branch = 'branch "master"'
		if not config.has_section(section_branch):
			config.add_section(section_branch)
		config.set(section_branch, "remote", remote_name)
		config.set(section_branch, "merge", merge)

	def _remote_server_name(self, remote_name):
		"""
		Get server name form git-remote identifier.
		"""
		remote = self._repo.remotes[remote_name]
		url = remote.config_reader.get("url")
		server_name = re.search('@(.+?)/', url).group(1)
		return server_name

	@need_initialized
	def remote(self, server=None, user="sglserver", branch="master", remote_name="server"):
		"""
		Set up a synchronization with remote server and return an updated list of servers.
		"""
		# set up a new remote server
		if server is not None:
			branch = "%s/%s" % (utils.hostname(), branch)
			url = "ssh://%s@%s/~" % (user, server)
			self._remote_set_server(remote_name, url, branch)

		# list all remote servers
		servers = {}
		for remote in self._repo.remotes:
			# TODO: temporary workaround, this seems to be some kind of race-condition
			#server_name = self._remote_server_name(remote)
			server_name = "server"
			servers[remote.name] = server_name

		return servers

	@need_initialized
	def sync(self, remote_name="server"):
		"""
		Run synchronization with remote server.
		"""
		# remote server is not set
		if len(self._repo.remotes) == 0:
			return False

		# fetch changes
		remote = self._repo.remotes[0]
		try:
			remote_info = remote.fetch()[0]
		except AssertionError:
			remote.push(force=True)
			return True

		remote_commit = remote_info.ref.commit
		local_commit = self._repo.active_branch.commit

		# no synchronization needed
		if local_commit == remote_commit:
			return True

		# merge changes, if there are some commits reachable from server, but not from client branch
		if len(self._repo.git.rev_list(remote.refs, "^master", max_count=1)) >= 0:
			conflict_strategy = "ours"
			if config.Config().get("sync.conflict") == "remote":
				conflict_strategy = "theirs"

			server_name = self._remote_server_name(remote_name)

			self._repo.git.merge(remote_info.ref.name, m="Merge changes from %s" % server_name,
				no_ff=True, strategy="recursive", strategy_option=conflict_strategy)

		# push the changes onto server
		remote.push(force=True)
		return True
