import fileinfo.attributes

class DiffHelper(object):
	def __init__(self, repository):
		self._repo = repository

	def _git2sgl_parse(self, headers, diffs, input):
		assert type(headers) is list
		assert type(diffs) is list

		# diff reading FSM:
		#
		# (log-header) --> (diff-header) --> (git-headers) --> (patch-content) --\
		#                       '\_______________________________________________/
		#
		# - new state has to be set using 'new_state' variable
		# - 'continue' will cause processing of the same line in the next loop

		state = None
		new_state = "log-header"
		diff = None

		lines = input.splitlines()
		if len(lines) > 0:
			line = lines.pop(0)
		else:
			line = None

		while line is not None:
			state = new_state

			if state == "log-header":
				if line.startswith("diff"):
					new_state = "diff-header"
					continue
				else:
					headers.append(line)

			elif state == "diff-header":
				# diff --git a/path/file b/path/file
				filename = line.split()[2][1:]
				diff = Diff(filename)
				if self._repo._allowed_file(filename):
					diffs.append(diff)
				new_state = "git-headers"

			elif state == "git-headers":
				if line.startswith("---"):
					pass
				elif line.startswith("+++"):
					new_state = "patch-content"
				else:
					diff.git_headers.append(line)

			elif state == "patch-content":
				if line.startswith("diff"):
					new_state = "diff-header"
					continue
				else:
					diff.content.append(line)

			else:
				assert False, "Unexpected FSM state '%s'." % state

			if len(lines) > 0:
				line = lines.pop(0)
			else:
				line = None

	def _git2sgl_add_nongit(self, diffs, old_attrs, new_attrs, limit_files):
		assert type(diffs) is list
		assert type(old_attrs) is fileinfo.attributes.AttributeSet
		assert type(new_attrs) is fileinfo.attributes.AttributeSet

		all_files = set(old_attrs.files()) & set(new_attrs.files())
		for file in all_files:
			if file in map(lambda diff: diff.filename, diffs):
				continue
			if limit_files is not None and file not in limit_files:
				continue
			elif (
				not file in old_attrs.files() or
				not file in new_attrs.files() or
				old_attrs[file] != new_attrs[file]
			):
				diffs.append(Diff(file))

	def git2sgl(self, input, old_attrs, new_attrs, limit_files=None):
		"""
		Convert diff in Git format to Sandglass format. In fact, filemode, owner and context is added.

		git header:
			diff --git a/path/file b/path/file
			index 0123456..789abcd 100644
			--- a/path/file
			+++ b/path/file

		sgl header:
			diff --sgl a/path/file b/path/file
			index 0123456..789abcd 100644
			mode 0644 0644
			owner root:root root:root
			context unconfined_u:object_r:etc_t:s0 unconfined_u:object_r:etc_t:s0
			--- a/path/file
			+++ b/path/file
		"""
		headers = []
		diffs = []

		self._git2sgl_parse(headers, diffs, input)
		self._git2sgl_add_nongit(diffs, old_attrs, new_attrs, limit_files)

		for diff in diffs:
			new = new_attrs[diff.filename]
			old = old_attrs[diff.filename]
			sgl_headers = fileinfo.attributes.Attribute.gen_diff_headers(old, new)
			diff.sgl_headers.extend(sgl_headers)

		headers_str = "\n".join(headers)
		diffs_str = "\n".join(map(lambda diff: diff.str(), sorted(diffs)))
		if len(headers_str) == 0:
			return diffs_str
		else:
			return "%s\n%s" % (headers_str, diffs_str)

class Diff(object):
	def __init__(self, filename):
		self.filename = filename
		self.git_headers = []
		self.sgl_headers = []
		self.content = []

	def __cmp__(self, other):
		assert type(self) == type(other)
		return cmp(self.filename, other.filename)

	def str(self):
		header = []
		header.append("diff --sgl a%(file)s b%(file)s" % { "file" : self.filename })
		header.extend(self.git_headers)
		header.extend(self.sgl_headers)
		header.append("--- a%s" % self.filename)
		header.append("+++ b%s" % self.filename)

		return "\n".join(header + self.content)
