"""
Storing file attributes (permissions, owner, SELinux context).
"""

import changes
from sandglass import distinfo

import os
import stat
import pwd
import grp
import selinux

EMPTY_FIELD = "-"

class AttributeSet(object):
	def __init__(self, file_list=None):
		if file_list is None:
			file_list = []

		self._list = dict(map(self._attribute_init, file_list))
		self.refresh()

	def _attribute_init(self, file):
		return (file, Attribute(file))

	def refresh(self):
		for file in sorted(self._list):
			self._list[file].load()

	def save(self, filename):
		with open(filename, "w") as f:
			for file in sorted(self._list):
				print >>f, self._list[file].to_string()

	def load(self, filename):
		self._list = {}
		with open(filename, "r") as f:
			self.load_s(f.read())

	def load_s(self, data):
		self._list = {}
		for line in data.splitlines():
			item = Attribute.from_string(line)
			self._list[item.filename] = item

	def files(self):
		return self._list.keys()

	def __getitem__(self, key):
		if key in self._list:
			return self._list[key]
		else:
			return Attribute(key)

class Attribute(object):
	"""
	Representation of attributes of one file.

	Stores file name, owning package, ownerships, permissions and SELinux context.
	"""

	def __init__(self, filename, mode = 0, user = None, group = None, context = None, package = None):
		self.filename = filename
		self.package = package
		self.mode = int(mode)
		self.user = user
		self.group = group
		self.context = context

		if package is not None:
			self.package = package
		else:
			self.package = distinfo.get_instance().file_package(self.filename)

	def __str__(self):
		return "Attribute('%s')" % self.filename

	def __eq__(self, other):
		return (
			self.mode == other.mode and self.user == other.user and
			self.group == other.group and self.context == other.context
		)

	def __ne__(self, other):
		return not (self == other)

	def diff(self, other):
		"""
		Get difference between two file attributes. Both Attribute objects have to refer to the same file.
		Return appropriate changes.Change object if the attributes differ. Otherwise return None.
		"""
		assert self.filename == other.filename

		if self.mode != other.mode:
			return changes.Mode(self.mode, other.mode)
		elif self.user != other.user or self.group != other.group:
			return changes.Owner([self.user, self.group], [other.user, other.group])
		elif self.context != other.context:
			return changes.Context(self.context, other.context)
		else:
			return None

	@classmethod
	def _filter_to_string(cls, value):
		"""Convert None attribute to string representing empty value."""
		if value is None:
			return EMPTY_FIELD
		else:
			return value

	@classmethod
	def _filter_from_string(cls, value):
		"""Convert string to attribute value considering string representing empty value."""
		if value == EMPTY_FIELD:
			return None
		else:
			return value

	def to_string(self):
		"""Convert Attribute object to string representation."""
		columns = map(Attribute._filter_to_string,
				[self.filename, self.package, self.mode, self.user, self.group, self.context])
		return " ".join(map(str, columns))

	@classmethod
	def from_string(cls, data):
		"""Convert string to Attribute object."""
		(filename, package, mode, user, group, context) = map(cls._filter_from_string, data.split())
		object = Attribute(filename, int(mode), user, group, context, package=package)
		return object

	def load(self):
		"""Read mode, permission and context of the file."""
		try:
			st = os.lstat(self.filename)

			self.mode = stat.S_IMODE(st.st_mode)
			self.user = pwd.getpwuid(st.st_uid).pw_name
			self.group = grp.getgrgid(st.st_gid).gr_name
		except OSError:
			self.mode = 0
			self.user = None
			self.group = None
		try:
			(len, self.context) = selinux.lgetfilecon(self.filename)
		except OSError:
			self.context = None

	def save(self, filename=None):
		"""Apply mode, permissions and context to the file."""
		if filename is None:
			filename = self.filename

		st = os.lstat(filename)

		# file mode (makes no sense for symlinks)
		used_mode = stat.S_IMODE(st.st_mode)
		if not os.path.islink(filename) and used_mode != self.mode:
			os.chmod(filename, self.mode)

		# owner
		used_uid = st.st_uid
		used_gid = st.st_gid
		uid = pwd.getpwnam(self.user).pw_uid
		gid = grp.getgrnam(self.group).gr_gid
		if used_uid != uid or used_gid != gid:
			os.lchown(filename, uid, gid)

		# selinux context
		if self.context is None and selinux.is_selinux_enabled() == 1:
			try:
				# matchpathcon reads only file format bits of mode
				(len, context) = selinux.matchpathcon(filename, st.st_mode)
			except OSError:
				context = None
		else:
			context = self.context

		try:
			(len, used_context) = selinux.lgetfilecon(filename)
			if context is not None and used_context != context:
				selinux.chcon(filename, context)
		except OSError:
			pass

	@classmethod
	def _gen_diff_headers_filter_mode(cls, mode):
		if mode == 0:
			return EMPTY_FIELD
		else:
			return "0%o" % mode

	@classmethod
	def _gen_diff_headers_filter_owner(cls, user, group):
		if user is None or group is None:
			return EMPTY_FIELD
		else:
			return "%s:%s" % (user, group)

	@classmethod
	def _gen_diff_headers_filter_context(cls, context):
		if context is None:
			return EMPTY_FIELD
		else:
			return context

	@classmethod
	def gen_diff_headers(cls, old, new):
		old_mode = cls._gen_diff_headers_filter_mode(old.mode)
		new_mode = cls._gen_diff_headers_filter_mode(new.mode)
		old_owner = cls._gen_diff_headers_filter_owner(old.user, old.group)
		new_owner = cls._gen_diff_headers_filter_owner(new.user, new.group)
		old_ctx = cls._gen_diff_headers_filter_context(old.context)
		new_ctx = cls._gen_diff_headers_filter_context(new.context)

		return [
			"mode %s %s" % (old_mode, new_mode),
			"owner %s %s" % (old_owner, new_owner),
			"context %s %s" % (old_ctx, new_ctx),
		]
