import time
import itertools
import sandglass.distinfo

class ChangeSet(object):
	"""
	The aggregation of changes (class Change).
	"""

	def __init__(self):
		self._changes = {}

	def __getitem__(self, key):
		return self._changes.__getitem__(key)

	def files(self):
		"""
		List of files whose changes this object holds.
		"""
		return sorted(self._changes.keys())

	def _find_change(self, list, change_type):
		"""
		Find object of a given type in the list of changes and return it. If no matching object is found,
		return None.
		"""
		for change in list:
			if type(change) is change_type:
				return change
		else:
			return None

	def add_or_update(self, file, new_change):
		"""
		Add another Change object to the list of changes performed on one file. Any old Change object
		of the same type (same base class) is removed before.
		"""
		assert new_change is not None

		self._changes.setdefault(file, [])
		change = self._find_change(self._changes[file], type(new_change))
		if change is not None:
			self._changes[file].remove(change)

		self._changes[file].append(new_change)

	def add_or_update_list(self, file, changes):
		"""
		Replace the list of all changes performed on one file.
		"""
		assert type(changes) is list
		self._changes[file] = changes

	def _merge_old_list(self, new_list, old_list):
		"""
		Merge changes from 'old_list' to 'new_list' (lists of Change objects).

		This action is performed by calling 'merge' method on objects matching the same type.
		"""
		for new in new_list:
			change_type = type(new)
			old = self._find_change(old_list, change_type)
			if old is not None:
				new.merge_old(old)

	def merge_old(self, old):
		"""
		Merge this object with another one. The changes stored in the second object are considered old.
		It might be necessary to confront the state of the changes with the old state. The action is
		performed by calling 'merge' method on matching objects (of Change type).
		"""
		for file in self._changes:
			if file in old._changes:
				self._merge_old_list(self._changes[file], old._changes[file])

	def time_first_change(self):
		"""Get timestamp of the first change in the list of changes."""
		when = None
		for change in itertools.chain(*self._changes.itervalues()):
			if when is None or when > change.first_change:
				when = change.first_change
		return when

	def time_last_change(self):
		"""Get timestamp of the last change in the list of changes."""
		when = None
		for change in itertools.chain(*self._changes.itervalues()):
			if when is None or when < change.last_change:
				when = change.last_change
		return when

	def by_package(self):
		"""
		Group changes by package name. Return a dictionary with package names as keys and ChangeSet objects
		as values.
		Shallow copies of the lists are performed.
		"""
		pkg_changesets = {}
		info = sandglass.distinfo.get_instance()
		for file in self._changes:
			package = info.file_package(file)
			pkg_changesets.setdefault(package, ChangeSet())
			pkg_changesets[package].add_or_update_list(file, self._changes[file])
		return pkg_changesets

class Change(object):
	"""
	Base class for any kind of change held on filesystem.
	"""
	def __init__(self, last_change = None):
		if last_change is None:
			last_change = time.time()

		self.last_change = last_change
		self.first_change = last_change

	def name(self):
		"""Return a string with the name of the change."""
		raise NotImplementedError()

	def detailed_str(self):
		"""Return a string with detailed information about the change."""
		raise NotImplementedError()

	def merge_old(self, old):
		"""Merge the old change into new one (self)."""
		self.first_change = old.first_change

class Modify(Change):
	"""
	Information about the change in a file.
	"""
	def __init__(self, added = 0, deleted = 0, mtime = None):
		Change.__init__(self, mtime)
		self.added = added
		self.deleted = deleted

	def name(self):
		return "modification"

	def detailed_str(self):
		return "lines: %d added, %d deleted" % (self.added, self.deleted)

class Delete(Change):
	"""
	Information about the file deletion.
	"""
	def __init__(self):
		Change.__init__(self)

	def name(self):
		return "delete"

	def detailed_str(self):
		return None

class Mode(Change):
	"""
	Information about the change of the file mode.
	"""
	def __init__(self, old = 0, new = 0):
		Change.__init__(self, None)
		self.old_mode = old
		self.new_mode = new

	def name(self):
		return "mode"

	def detailed_str(self):
		return "0%o -> 0%o" % (self.old_mode, self.new_mode)

	def merge_old(self, old):
		Change.merge_old(self, old)
		if self.new_mode == old.new_mode:
			self.last_change = old.last_change

class Owner(Change):
	"""
	Information about the change of the file ownership.
	"""
	def __init__(self, old = [None, None], new = [None, None]):
		"""
		Parameters 'old' and 'new' are lists with two values, user and group.
		"""
		Change.__init__(self, None)
		self.old_owner = old
		self.new_owner = new

	def name(self):
		return "owner"

	def detailed_str(self):
		return "%s:%s -> %s:%s" % (self.old_owner[0], self.old_owner[1], self.new_owner[0], self.new_owner[1])

	def merge_old(self, old):
		Change.merge_old(self, old)
		if self.new_owner == old.new_owner:
			self.last_change = old.last_change

class Context(Change):
	"""
	Information about the change of the file SELinux context.
	"""
	def __init__(self, old, new):
		Change.__init__(self, None)
		self.old_context = old
		self.new_context = new

	def name(self):
		return "context"

	def detailed_str(self):
		return "%s -> %s" % (self.old_context, self.new_context)

	def merge_old(self, old):
		Change.merge_old(self, old)
		if self.new_context == old.new_context:
			self.last_change = old.last_change
