import clientrepository
import serverrepository
import exceptions
import distinfo
import utils
import os.path
import pwd
import grp

DEFAULT_DATA_DIR = "/var/lib/sandglass"
DEFAULT_ROOT_DIR = "/"
DEFAULT_SERVER_OWNER = "sglserver:sglserver"

class SandglassBase(object):
	def _delegate_method(self, instance, method_name):
		"""
		Creates instance method which will be delegated to other object (delegee).
		"""
		if hasattr(self, method_name):
			raise AttributeError("Method already defined.")

		delegee_method = getattr(instance, method_name)
		def delegator(*args, **kwargs):
			return delegee_method(*args, **kwargs)
		setattr(self, method_name, delegator)


class Sandglass(SandglassBase):
	def __init__(self, **kwargs):
		"""
		Initializes Sandglass object.
		"""
		if not kwargs.has_key("data_dir") or kwargs["data_dir"] is None:
			kwargs["data_dir"] = DEFAULT_DATA_DIR
		if not kwargs.has_key("root_dir") or kwargs["root_dir"] is None:
			kwargs["root_dir"] = DEFAULT_ROOT_DIR

		self._data_dir = kwargs["data_dir"]
		self._root_dir = kwargs["root_dir"]
		self._repo = clientrepository.ClientRepository(**kwargs)
		self._dist_info = distinfo.get_instance()

		for method in ["init", "is_initialized", "changed", "tracked", "log", "show", "diff",
							"dump", "recover", "commit", "remote", "sync"]:
			self._delegate_method(self._repo, method)

	def add(self, files, config_only = True):
		"""
		Add files or directories to the repository.

		If 'config_only' is set to True, only configuration files (according to
		package manager database) will be added and the other will be ignored.
		The setting of this option does not have effect on adding files.
		"""

		files_add = []
		for path in files:
			if utils.is_dir(path):
				self._add_dir(files_add, path, config_only)
			elif os.path.exists(path): # file or symlink
				files_add.append(path)
			else:
				raise exceptions.IOError("Path '%s' not found." % path)

		self._repo.add(files_add)

	def add_changed(self, files):
		if len(files) == 0:
			files = self.changed().files()
		if len(files) == 0:
			raise exceptions.InvalidOperation("There are no changes to be commited.")
		self._repo.add(files)

	def _add_dir(self, list, dirname, config_only):
		if config_only:
			files = self._dist_info.config_files(prefix = dirname)
			if len(files) == 0:
				raise exceptions.InvalidOperation("There are no configuration files in '%s'." % dirname)
		else:
			files = utils.files_from_dir(dirname)

		list.extend(files)

	def rm(self, files):
		"""
		Remove files from the list of versioned files.
		"""
		self._repo.untrack(files)

	def get_package_files(self, package):
		"""
		Return list of configuration files owned by given package.
		"""
		return self._dist_info.package_files(package)

class SandglassServer(SandglassBase):

	def __init__(self, data_dir = None, owner = DEFAULT_SERVER_OWNER):
		super(SandglassServer, self).__init__()

		if data_dir is None: data_dir = DEFAULT_DATA_DIR
		if owner is None: owner = DEFAULT_SERVER_OWNER

		(uid, gid) = self._owner_uid_gid(owner)
		self._repo = serverrepository.ServerRepository(data_dir, uid, gid)

		for method in ["init", "status", "log", "show", "diff", "delete", "pick"]:
			self._delegate_method(self._repo, method)

	def _owner_uid_gid(self, owner):
		(username, group) = owner.split(":")
		uid = pwd.getpwnam(username).pw_uid
		gid = grp.getgrnam(group).gr_gid
		return (uid, gid)
