import exceptions
import fileinfo
import diff
import os.path
import git.objects.commit
import functools
import inspect

def need_initialized(method):
	"""
	Decorator function. Raise exception on method call if the repository is not initialized.
	"""
	assert inspect.isfunction(method)
	@functools.wraps(method)
	def inner(self, *args, **kwargs):
		if self.is_initialized():
			return method(self, *args, **kwargs)
		else:
			raise exceptions.RepositoryInitialization("Repository is not initialized.")
	return inner

def need_not_initialized(method):
	"""
	Decorator function. Raise exception on method call if the repository is initialized.
	"""
	assert inspect.isfunction(method)
	@functools.wraps(method)
	def inner(self, *args, **kwargs):
		if not self.is_initialized():
			return method(self, *args, **kwargs)
		else:
			raise exceptions.RepositoryInitialization("Repository is already initialized.")
	return inner

class Repository(object):
	"""
	This is a base class for ClientRepository and ServerRepository.
	"""

	def __init__(self):
		if type(self) is Repository:
			raise NotImplementedError()

		self._repo = None

		# repository and working tree location
		self._repo_dir = None
		self._root_dir = None

	def is_initialized(self):
		"""
		Return True if the repository is initialized.
		"""
		return self._repo is not None

	def _allowed_file(self, filename):
		"""
		Return True if the given file is not Sandglass internal.
		"""
		return not os.path.basename(filename).startswith(".sgl")

	def _get_commit(self, revision):
		"""
		Return Git commit object. Method accepts string (revision name) or other Git commit.
		"""
		if type(revision) is str:
			return self._repo.commit(revision)
		elif type(revision) is git.objects.commit.Commit:
			return revision
		else:
			raise exceptions.InvalidOperation("Invalid type of 'revision' argument.")

	def get_attributes(self, revision):
		"""
		Load file attributes from repository at a given commit.
		"""
		commit = self._get_commit(revision)
		content = commit.tree[".sglattributes"].data
		attrs = fileinfo.attributes.AttributeSet()
		attrs.load_s(content)
		return attrs

	def repo_filename(self, real_filename):
		"""
		Convert real filename to repository filename. Opposite of repo_filename.
		"""
		# has to be overriden in derived classes
		raise NotImplementedError()

	def real_filename(self, repo_filename):
		"""
		Convert repository filename to real filename. Opposite of repo_filename.
		"""
		# has to be overriden in derived classes
		raise NotImplementedError()

	@need_initialized
	def log(self, revision=None, files=None):
		"""
		Get commit history. The history can be limited by passing a list of files.
		"""
		files = filter(self._allowed_file, map(self.repo_filename, files))
		return self._repo.git.log(revision, "--", files, pretty="format:%h %ci - %s")

	@need_initialized
	def show(self, commit, files=None):
		"""
		Show content of given commit (as a diff). The output can be limited by passing a list of files.
		"""
		diff_helper = diff.DiffHelper(self)
		if files is None:
			files = []

		commit = self._get_commit(commit)
		new_attrs = self.get_attributes(commit)
		if len(commit.parents) > 0:
			old_attrs = self.get_attributes(commit.parents[0])
		else:
			old_attrs = fileinfo.attributes.AttributeSet()

		git_output = self._repo.git.show(commit, "--", files)
		sgl_output = diff_helper.git2sgl(git_output, old_attrs, new_attrs, limit_files=files)
		return sgl_output

	@need_initialized
	def diff(self, old_commit, new_commit, files=None):
		"""
		Show diff between two revisions. The output can be limited by passing a list of files.
		"""
		diff_helper = diff.DiffHelper(self)
		files = filter(self._allowed_file, map(self.repo_filename, files))

		old_attrs = self.get_attributes(old_commit)
		new_attrs = self.get_attributes(new_commit)

		git_output = self._repo.git.diff(old_commit, new_commit, "--", files)
		sgl_output = diff_helper.git2sgl(git_output, old_attrs, new_attrs, limit_files=files)
		return sgl_output

	@need_initialized
	def dump(self, filename, commit):
		"""
		Get file content at a given commit.
		"""
		commit = self._get_commit(commit)
		filename = self.repo_filename(filename)
		return commit.tree[filename].data
