import os
import shutil
import stat
import pwd
import re

import git
import repository
import override
import exceptions
import utils
import fileinfo
import diff
import cache

need_initialized = repository.need_initialized
need_not_initialized = repository.need_not_initialized

class ServerRepository(repository.Repository):
	def __init__(self, data_dir, uid, gid):
		self._repo = None
		self._repo_dir = os.path.join(data_dir, "server")
		self._owner_uid = uid
		self._owner_gid = gid

		self._setup()

	def _setup(self):
		# not initialized
		if not os.path.exists(os.path.join(self._repo_dir, "objects")):
			return False

		# initialization check
		repo_stat = os.lstat(self._repo_dir)
		if not stat.S_ISDIR(repo_stat.st_mode):
			raise exceptions.RepositoryInitialization("Repository path '%s' has to be a directory." % self._repo_dir)
		elif repo_stat.st_uid != self._owner_uid or repo_stat.st_gid != self._owner_gid:
			raise exceptions.RepositoryInitialization("Repository has invalid ownership.")

		self._repo = override.Repo(self._repo_dir, None)
		return True

	def _run_as_owner(self, method, *args, **kwargs):
		"""
		Runn passed method/function with server repository owner permissions.
		"""
		saved_exception = None
		try:
			os.setegid(self._owner_gid)
			os.seteuid(self._owner_uid)
			method(*args, **kwargs)
		except Exception as e:
			saved_exception = e
		finally:
			os.seteuid(os.getuid())
			os.setegid(os.getgid())
		if saved_exception is not None:
			raise saved_exception

	def real_filename(self, repo_filename):
		return "/%s" % repo_filename

	def repo_filename(self, real_filename):
		return real_filename.lstrip("/")

	@need_not_initialized
	def _init_repo(self):
		new_repo = override.Repo.init(self._repo_dir, bare = True)

		config = new_repo.config_writer()
		hostname = utils.hostname()
		username = pwd.getpwuid(self._owner_uid).pw_name
		config.add_section("user")
		config.set("user", "name", "%s (%s)" % (username, hostname))
		config.set("user", "email", "%s@%s" % (username, hostname))

	def init(self, force = False):
		"""
		Initialize server repository. If the repository is already initialized, 'force' has to be set to True.
		"""
		if force:
			# whole directory must not be removed, as there might be some ssh keys installed
			for filename in filter(lambda f: not f.startswith("."), os.listdir(self._repo_dir)):
				filename = os.path.join(self._repo_dir, filename)
				if os.path.isdir(filename):
					shutil.rmtree(filename, True)
				else:
					os.unlink(filename)
			self._repo = None

		self._run_as_owner(self._init_repo)
		self._setup()

	def status(self):
		"""
		Return a dictionary with information about synchronized clients.

		Example result:
		{ "alice": ("1234567", "2011-01-01 01:01:01"), "bob": ("abcdefg", "2011-02-02 02:02:02") }
		"""
		result = {}
		for branch in self._repo.branches:
			commit = branch.commit
			client = re.match('(.+)/master$', branch.name).group(1)
			revision = str(branch.commit)[:7]
			date = utils.timestamp_to_str(commit.committed_date)
			result[client] = (revision, date)
		return result

	def _get_client_revision(self, client_name, branch="master"):
		"""
		Get revision name by a given client name.
		"""
		return "refs/heads/%s/%s" % (client_name, branch)

	@need_initialized
	def log(self, client, files = None):
		"""
		Show client history.
		"""
		client_rev = self._get_client_revision(client)
		try:
			return super(ServerRepository, self).log(client_rev, files)
		except git.GitCommandError as e:
			raise exceptions.InvalidRevision("Client does not exist.")

	@need_initialized
	def delete(self, client):
		"""
		Delete client history.
		"""
		# delete the directory directly, git doesn't want to remove the branch in bare repository
		client_dir = os.path.join(self._repo_dir, "refs", "heads", client)
		if not os.path.exists(client_dir):
			raise exceptions.InvalidRevision("Client does not exist.")
		shutil.rmtree(client_dir)

	@need_initialized
	def pick(self, commit, client):
		"""
		"""
		raise NotImplementedError()
