import os.path
import os
import socket
import atexit
import tempfile
from subprocess import Popen, PIPE
import time
import datetime

SECURE_UMASK = 0077

class Command(object):
	def __init__(self, args):
		"""Execute command and store it's exit code and output.

		   cmd = Command(["ls", "-alF"])
		   if cmd.exit_code != 0:
		       print cmd.stderr
		"""
		self._proc = Popen(args, stdout=PIPE, stderr=PIPE)
		(self._stdout, self._stderr) = self._proc.communicate()

	@property
	def exit_code(self):
		"""Process exit code."""
		return self._proc.returncode

	@property
	def stdout(self):
		"""Process output."""
		return str(self._stdout)

	@property
	def stderr(self):
		"""Process error output."""
		return str(self._stderr)

def abs_path(path):
	"""Convert path to absolute and normalized path."""
	full = os.path.join(os.getcwd(), path)
	return os.path.normpath(full)

def is_dir(path):
	"""Return True if specified path is real directory (not symlink), False otherwise."""
	return os.path.isdir(path) and not os.path.islink(path)

def files_from_dir(path):
	"""Create list of files in given directory (recursively)."""
	files_add = []
	for (root, dir, files) in os.walk(path, followlinks = False):
		files_add.extend(map(lambda f: os.path.join(root, f), files))

	return files_add

def hostname():
	name = socket.getfqdn()
	if name == "localhost":
		name = socket.gethostname()

	return name

def file_mtime(filename, must_exist=True):
	try:
		return os.lstat(filename).st_mtime
	except OSError as e:
		if must_exist:
			raise e
		else:
			return 0.0

def temp_file():
	"""
	Return the name of new temporary file created by tempfile.mkstemp(). The file will be automatically deleted
	at program termination.
	"""
	(fd, filename) = tempfile.mkstemp()
	atexit.register(_temp_file_delete, filename)
	os.close(fd)
	return filename

def _temp_file_delete(filename):
	if os.path.exists(filename):
		os.unlink(filename)

def secure_open(filename, *args, **kwargs):
	"""
	Open file handle. If the file is created, set it user-accessible only.
	Works the same as Python builtin 'open' function
	"""
	old_umask = os.umask(SECURE_UMASK)
	fh = open(filename, *args, **kwargs)
	os.umask(old_umask)
	return fh

def secure_makedirs(*args, **kwargs):
	"""
	Create directories recursively. Newly created directories are set user-accessible only.
	Works the same as Python 'os.makedirs' function.
	"""
	old_umask = os.umask(SECURE_UMASK)
	result = os.makedirs(*args, **kwargs)
	os.umask(old_umask)
	return result

def time_remains(timestamp):
	now = time.time()
	return int(timestamp - now)

def time_ago(timestamp):
	return -1 * time_remains(timestamp)

def time_relative(timestamp):
	diff = time_remains(timestamp)
	if diff == 0:
		return "now"
	elif diff > 0:
		format = "in %d %s"
	else:
		format = "%d %s ago"

	diff = abs(diff)
	if diff < 60:
		unit = ["second", "seconds"]
	elif diff < 3600:
		diff /= 60
		unit = ["minute", "minutes"]
	elif diff < 86400:
		diff /= 3600
		unit = ["hour", "hours"]
	else:
		diff /= 86400
		unit = ["day", "days"]

	if diff == 1:
		unit = unit[0]
	else:
		unit = unit[1]

	return format % (diff, unit)

def timestamp_to_str(timestamp, format=None):
	if format is None:
		format = "%Y-%m-%d %H:%M:%S"
	d = datetime.datetime.fromtimestamp(timestamp)
	return d.strftime(format)
