# http://yum.baseurl.org/wiki/WritingYumPlugins

import os
import sandglass.cache
import yum.plugins
import yum.constants

requires_api_version = "2.6"
plugin_type = (yum.plugins.TYPE_CORE, yum.plugins.TYPE_INTERACTIVE)

sgl = None
sgl_config = None
sgl_enabled = False
sgl_post_operations = []

class SglYumOperation(object):
	def __init__(self, package, files, operation, info = None):
		self.package = package
		self.files = files
		self.operation = operation
		self.info = info

	def perform(self, sgl):
		if len(self.files) > 0:
			self.operation(self.files)

	def get_message(self):
		if self.info is not None:
			return "@%s (%s)" % (self.package, self.info)
		else:
			return "@%s" % self.package

def sgl_transaction_pkgs(transaction, states):
	"""Return set with the names of the packages in a given transaction having a given state."""
	names = [txm.name for txm in transaction.getMembersWithState(output_states = states)]
	return set(names)

def sgl_parse_transaction(transaction):
	"""Return tuple (install, erase, change) where each element is a list with package names."""
	install = sgl_transaction_pkgs(transaction, yum.constants.TS_INSTALL_STATES)
	erase = sgl_transaction_pkgs(transaction, yum.constants.TS_REMOVE_STATES)

	change = install & erase
	install = install - change
	erase = erase - change

	return (install, erase, change)

def sgl_commit(conduit, operations, stage = "yum"):
	"""Perform a commit. List 'operations' is a list of SglYumOperation objects."""
	if len(operations) == 0:
		return

	huge_commit = sgl_config.get("yum.commit_granularity", "package") == "update"
	if huge_commit and len(operations) == 1:
		huge_commit = False

	for operation in operations:
		operation.perform(sgl)
		if not huge_commit:
			conduit.info(2, "Performing Sandglass autocommit for package '%s'." % operation.package)
			sgl.commit("%s autocommit: %s" % (stage, operation.get_message()))

	if huge_commit:
		conduit.info(2, "Performing Sandglass autocommit (%d packages)." % len(operations))
		commit_details = "\n".join(map(lambda op: " - %s" % op.get_message(), operations))
		sgl.commit("%s autocommit: %d packages\n\n%s" % (stage, len(operations), commit_details))

def sgl_operations_add_changed(operations, packages_tx):
	changes = sgl.changed().by_package()
	packages_commit = filter(changes.__contains__, packages_tx)
	for package in packages_commit:
		op = SglYumOperation(package, changes[package].files(), sgl.add)
		operations.append(op)

def sgl_operations_add_erased(operations, packages_tx):
	if len(packages_tx) == 0:
		return

	files_tracked = set(sgl.tracked())
	for package in packages_tx:
		files_pkg = set(sgl.get_package_files(package))
		files = files_tracked & files_pkg
		if len(files) > 0:
			op = SglYumOperation(package, list(files), sgl.rm, "erase")
			operations.append(op)

def sgl_operations_add_new(operations, packages_tx):
	if not sgl_config.getboolean("yum.autotracking", True):
		return

	for package in packages_tx:
		files = sgl.get_package_files(package)
		op = SglYumOperation(package, files, sgl.add, "install")
		operations.append(op)

def init_hook(conduit):
	global sgl, sgl_enabled, sgl_config

	# will not work without root permissions
	if os.geteuid() != 0:
		return

	sgl = sandglass.Sandglass()
	if not sgl.is_initialized():
		conduit.info(2, "Sandglass repository not initialized.")
		return

	sgl_config = sandglass.config.get_instance()
	if not sgl_config.getboolean("yum.enabled", True):
		conduit.info(2, "Integration of Sandglass into Yum is disabled.")
		return

	sgl_enabled = True

def pretrans_hook(conduit):
	if not sgl_enabled:
		return

	(install, erase, change) = sgl_parse_transaction(conduit.getTsInfo())

	operations = []
	sgl_operations_add_changed(operations, (change | erase))
	sgl_operations_add_erased(sgl_post_operations, erase)

	sgl_commit(conduit, operations, "yum(pre)")

def posttrans_hook(conduit):
	if not sgl_enabled:
		return

	(install, erase, change) = sgl_parse_transaction(conduit.getTsInfo())

	operations = []
	sgl_operations_add_changed(operations, change)
	sgl_operations_add_new(operations, install)
	operations.extend(sgl_post_operations)

	sgl_commit(conduit, operations, "yum(post)")
